import smtplib
from email.message import Message
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Optional

from .strip_html import strip_html

try:
    import aiosmtplib
except ImportError:
    aiosmtplib = None

try:
    import jinja2
except ImportError:
    jinja2 = None


class TechsolMailer(object):
    """
    A base mailer class that stores SMTP server info.

    Args:
        hostname: Server name (or IP) to connect to.
        port: Server port.
        username: Username to login as after connect.
        password: Password for login after connect.
        use_tls: If True, make the initial connection to the server over TLS/SSL. Note that if the server supports STARTTLS only, this should be False.
        start_tls: If True, make the initial connection to the server over plaintext, and then upgrade the connection to TLS/SSL. Not compatible with ``use_tls``.

    Raises:
        ValueError: if specified incorrect options
    """

    def __init__(
        self,
        *,
        hostname: str,
        port: int,
        username: str,
        password: str,
        use_tls: bool = False,
        start_tls: bool = False,
    ):
        if use_tls and start_tls:
            raise ValueError("use_tls and start_tls are mutually exclusive.")

        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls
        self.start_tls = start_tls

    def send(self, message: Message):
        """
        Sends an email message using the configured SMTP server.
        """

        smtp_cls = smtplib.SMTP_SSL if self.use_tls else smtplib.SMTP
        smtp = smtp_cls(
            host=self.hostname,
            port=self.port,
        )
        if self.start_tls:
            smtp.starttls()
        smtp.ehlo()
        smtp.login(
            user=self.username,
            password=self.password,
        )
        smtp.sendmail(
            from_addr=message["From"],
            to_addrs=message.get_all("To", []),
            msg=message.as_string(),
        )
        smtp.quit()

    async def send_async(self, message: Message):
        """
        Sends an email message using the configured SMTP server.

        Raises:
            RuntimeError: if ``aiosmtplib`` is not installed
        """

        if aiosmtplib is None:
            raise RuntimeError("Async support is not installed. Install it with 'pip install techsol-mailer[async]'")

        await aiosmtplib.send(
            message,
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            use_tls=self.use_tls,
            start_tls=self.start_tls,
        )


class TemplatedMailer(TechsolMailer):
    """
    A :class:`TechsolMailer` with Jinja templates support.

    :meth:`~.send` and :meth:`~.send_async` methods work here too.

    Args:
        jinja_env: a :class:`jinja2.Environment` to get templates from
        default_sender: default message sender
        **kwargs: same as in :class:`TechsolMailer`

    Will raise :class:`RuntimeError` on init if ``jinja2`` is not installed.
    """

    def __init__(
        self,
        *,
        jinja_env: jinja2.Environment,
        default_sender: str = "MediaMath TechSol <noreply-techsol@mediamath.com>",
        **kwargs,
    ):
        if jinja2 is None:
            raise RuntimeError("Templating support is not installed. Install it with 'pip install techsol-mailer[templates]'")

        self.jinja_env = jinja_env
        self.default_sender = default_sender
        super().__init__(**kwargs)

    @classmethod
    def from_package(cls, package_name: str = "app", **kwargs):
        """
        Sets up a :class:`TemplatedMailer` for a package ``package_name``.

        Args:
            package_name: import name of the package that contains the template directory, defaults to ``app``
            default_sender: default message sender

        The rest are same as in :class:`TechsolMailer`.
        """

        if jinja2 is None:
            raise RuntimeError("Templating support is not installed. Install it with 'pip install techsol-mailer[templates]'")

        return cls(
            jinja_env=jinja2.Environment(
                loader=jinja2.PackageLoader(package_name),
                autoescape=jinja2.select_autoescape()
            ),
            **kwargs,
        )

    def render_email(
        self,
        *,
        subject: str,
        template_name: str,
        context: dict,
        recipient: str,
        sender: Optional[str] = None,
    ) -> MIMEMultipart:
        """
        Renders a Jinja template into a :class:`~email.mime.multipart.MIMEMultipart`
        message and returns it.

        Looks for both ``{template_name}.html`` and ``{template_name}.txt``
        templates. If only the HTML part is present, will strip resulting HTML
        of all formatting in order to get plain text version. Note that e. g.
        links won't be preserved this way.

        Args:
            subject: message subject
            template_name: template filename without suffix
            context: variables to pass into the template
            recipient: message recipient, either a bare email address or ``Name Surname <email@example.org>``
            sender: message sender
        """

        if jinja2 is None:
            raise RuntimeError("Templating support is not installed. Install it with 'pip install techsol-mailer[templates]'")

        message = MIMEMultipart("alternative")
        message["From"] = sender or self.default_sender
        message["To"] = recipient
        message["Subject"] = subject

        html_part = self.jinja_env.get_template(f"{template_name}.html").render(**context)
        try:
            plain_text_part = self.jinja_env.get_template(f"{template_name}.txt").render(**context)
        except jinja2.TemplateNotFound:
            plain_text_part = strip_html(html_part)

        message.attach(MIMEText(plain_text_part, "plain", "utf-8"))
        message.attach(MIMEText(html_part, "html", "utf-8"))
        return message

    def send_from_template(
        self,
        *,
        subject: str,
        template_name: str,
        context: dict,
        recipient: str,
        sender: Optional[str] = None,
    ):
        """
        A shorthand to render an email from template and send it using the
        configured server. Parameters are the same as in
        :meth:`~.render_email`.
        """
        message = self.render_email(
            subject=subject,
            template_name=template_name,
            context=context,
            recipient=recipient,
            sender=sender,
        )
        return self.send(message)

    async def send_from_template_async(
        self,
        *,
        subject: str,
        template_name: str,
        context: dict,
        recipient: str,
        sender: Optional[str] = None,
    ):
        """
        A shorthand to render an email from template and send it using the
        configured server. Parameters are the same as in
        :meth:`~.render_email`.

        Raises:
            RuntimeError: if aiosmtplib is not installed
        """

        message = self.render_email(
            subject=subject,
            template_name=template_name,
            context=context,
            recipient=recipient,
            sender=sender,
        )
        return await self.send_async(message)
